module Infer where

open import Data.Unit.Base using (⊤; tt)
open import Data.Nat hiding (erase)
open import Data.Nat.Properties using (n≤1+n; n≤m+n; ≤-step; 1+n≰n)
open import Data.Fin hiding (_+_; _≤_; _<_)
open import Data.Fin.Properties
open import Data.Vec
open import Data.Sum
open import Data.Product
open import Data.Maybe.Base
open import Data.Empty
open import Function using (_∘_)
open import Relation.Nullary
open import Relation.Binary using (DecTotalOrder)
open import Relation.Binary.PropositionalEquality
open ≡-Reasoning

open import Regular
open import Properties
open import Occur
open import Subst
open import Mgu
open import STLC

--------------------------------------------------------------------------------

private
  thin-inject≤ : {m m′′ : ℕ} → {y′ : Fin m′′} → {x : Fin m} →
          {m≤m′′ : m ≤ m′′} →
          thin (inject≤ (fromℕ m) (s≤s m≤m′′)) y′ ≡
          inject≤ (inject≤ x (n≤m+n 1 m)) (s≤s m≤m′′) →
          y′ ≡ inject≤ x m≤m′′
  thin-inject≤ {zero} {x = ()} eq
  thin-inject≤ {suc m} {y′ = zero} {zero} {s≤s m≤m′′} eq = refl
  thin-inject≤ {suc m} {y′ = suc y′} {zero} ()
  thin-inject≤ {suc n} {y′ = zero} {suc x} {s≤s m≤m′′} ()
  thin-inject≤ {suc n} {y′ = suc y′} {suc x} {s≤s m≤m′′} eq =
    cong suc (thin-inject≤ (suc-equal eq))

  m-less : {m : ℕ} → (x : Fin m) →
          ¬ (fromℕ m ≡ inject≤ x (n≤m+n 1 m))
  m-less {zero} () eq
  m-less {suc m} zero ()
  m-less {suc m} (suc x) eq = m-less x (suc-equal eq)

  ⟨skip-fresh-type⟩ : {G : Code} → {m m′ m′′ : ℕ} → (m≤m′′ : m ≤ m′′) →
          (f : Fin m′′ → Type m′) → (t′ : Type m′) →
          (d : ⟦ G ⟧ (Type m)) →
          (r : ⟦ G ⟧′ (λ t → 
            applyUnifier≤ (f-for f t′ (inject≤ (fromℕ m) (s≤s m≤m′′)))
                          (s≤s m≤m′′) (lift≤ (n≤m+n 1 m) t) ≡
            applyUnifier≤ f m≤m′′ t) d) →
          fmap {TypeF} G ⟨_⟩ (f-for f t′ (inject≤ (fromℕ m) (s≤s m≤m′′)))
            (fmap {TypeF} G ⟨_⟩ (λ x → ⟪ inject≤ x (s≤s m≤m′′) ⟫)
              (fmap G ⟨_⟩ (λ x → ⟪ inject≤ x (n≤m+n 1 m) ⟫) d)) ≡
          fmap {TypeF} G ⟨_⟩ f (fmap G ⟨_⟩ (λ x → ⟪ inject≤ x m≤m′′ ⟫) d)
  ⟨skip-fresh-type⟩ {U} m≤m′′ f t′ tt tt = refl
  ⟨skip-fresh-type⟩ {I} m≤m′′ f t′ d r = r
  ⟨skip-fresh-type⟩ {G₁ ⊕ G₂} m≤m′′ f t′ (inj₁ d) r =
    cong inj₁ (⟨skip-fresh-type⟩ {G₁} m≤m′′ f t′ d r)
  ⟨skip-fresh-type⟩ {G₁ ⊕ G₂} m≤m′′ f t′ (inj₂ d) r =
    cong inj₂ (⟨skip-fresh-type⟩ {G₂} m≤m′′ f t′ d r)
  ⟨skip-fresh-type⟩ {G₁ ⊗ G₂} m≤m′′ f t′ (d₁ , d₂) (r₁ , r₂) =
    cong₂ _,_ (⟨skip-fresh-type⟩ {G₁} m≤m′′ f t′ d₁ r₁)
              (⟨skip-fresh-type⟩ {G₂} m≤m′′ f t′ d₂ r₂)

  ⟪skip-fresh-type⟫ : {m m′ m′′ : ℕ} → (m≤m′′ : m ≤ m′′) →
          (f : Fin m′′ → Type m′) → (t′ : Type m′) → (x : Fin m) →
          applyUnifier≤ (f-for f t′ (inject≤ (fromℕ m) (s≤s m≤m′′)))
                        (s≤s m≤m′′) (lift≤ (n≤m+n 1 m) ⟪ x ⟫) ≡
          applyUnifier≤ f m≤m′′ ⟪ x ⟫
  ⟪skip-fresh-type⟫ {m} m≤m′′ f t′ x
    with thick₂ (inject≤ (fromℕ m) (s≤s m≤m′′))
               (inject≤ (inject≤ x (n≤m+n 1 m)) (s≤s m≤m′′))
  ... | inj₂ (y′ , eq) = cong f (thin-inject≤ eq)
  ... | inj₁ eq with inject≤-equal eq
  ... | eq′ = ⊥-elim (m-less x eq′)

  skip-fresh-type : {m m′ m′′ : ℕ} → (m≤m′′ : m ≤ m′′) →
          (f : Fin m′′ → Type m′) → (t′ : Type m′) → (t : Type m) →
          applyUnifier≤ (f-for f t′ (inject≤ (fromℕ m) (s≤s m≤m′′)))
                        (s≤s m≤m′′) (lift≤ (n≤m+n 1 m) t) ≡
          applyUnifier≤ f m≤m′′ t
  skip-fresh-type {m} m≤m′′ f t′ t =
    ind (λ t → 
          applyUnifier≤ (f-for f t′ (inject≤ (fromℕ m) (s≤s m≤m′′)))
                        (s≤s m≤m′′) (lift≤ (n≤m+n 1 m) t) ≡
          applyUnifier≤ f m≤m′′ t)
        (λ d r → cong ⟨_⟩ (⟨skip-fresh-type⟩ {TypeF} m≤m′′ f t′ d r))
        (λ x → ⟪skip-fresh-type⟫ m≤m′′ f t′ x)
        t

  skip-fresh : {m n m′ m′′ : ℕ} → (Γ : Cxt {m} n) → (m≤m′′ : m ≤ m′′) →
          (f : Fin m′′ → Type m′) → (t′ : Type m′) →
          applyFunCxt≤ (f-for f t′ (inject≤ (fromℕ m) (s≤s m≤m′′)))
                       (s≤s m≤m′′) (liftCxt 1 Γ) ≡
          applyFunCxt≤ f m≤m′′ Γ
  skip-fresh [] m≤m′′ f t′ = refl
  skip-fresh {m} (t ∷ Γ) m≤m′′ f t′ 
    = cong₂ _∷_ (skip-fresh-type m≤m′′ f t′ t) (skip-fresh Γ m≤m′′ f t′)

untypable : {m n : ℕ} → (Γ : Cxt {m} n) → (s : WellScoped n) → Set
untypable {m} Γ s =
        (l′′ l′ : ℕ) → (m≤l′′ : m ≤ l′′) → (f₀ : Fin l′′ → Type l′) →
        (t₀ : Type l′) → (w₀ : WellTyped (applyFunCxt≤ f₀ m≤l′′ Γ) t₀) →
        ¬ erase w₀ s

mgt : {m m′ m′′ n : ℕ} → (σ : AList TypeF m′′ m′) →
        (Γ : Cxt {m} n) → (s : WellScoped n) → (t : Type m′) → Set
mgt {m} {m′} {m′′} σ Γ s t = 
        (l′′′ l′ : ℕ) → (m≤l′′′ : m ≤ l′′′) →
        (f₀ : Fin l′′′ → Type l′) → (t₀ : Type l′) →
        (w₀ : WellTyped (applyFunCxt≤ f₀ m≤l′′′ Γ) t₀) →
        erase w₀ s →
        Σ[ k′′ ∈ ℕ ] Σ[ k′′′ ∈ ℕ ] Σ[ m′≤k′′ ∈ m′ ≤ k′′ ]
        Σ[ m′′≤k′′′ ∈ m′′ ≤ k′′′ ] Σ[ para ∈ m′′≤k′′′ // m′≤k′′ ]
        Σ[ m≤k′′′ ∈ m ≤ k′′′ ] Σ[ f′ ∈ (Fin k′′ → Type l′) ] 
        (((x : Fin m) →
          f₀ (inject≤ x m≤l′′′) ≡
          (f′ +⟨ para ⟩′ (sub σ)) (inject≤ x m≤k′′′)) ×
         t₀ ≡ applyUnifier≤ f′ m′≤k′′ t)

private
  bot-f : {m n : ℕ} → suc n ≤ m → m ≡ n → ⊥
  bot-f 1+n≤m m≡n rewrite m≡n = 1+n≰n 1+n≤m

  thinmy′≡y→y′≡y : {m : ℕ} → {y y′ : Fin m} →
          thin (fromℕ m) y′ ≡ inject≤ y (n≤m+n 1 m) → y′ ≡ y
  thinmy′≡y→y′≡y {y′ = y′} thinmy′≡y with thin-max y′
  ... | eq rewrite thinmy′≡y = sym (inject≤-equal eq)

f-for-xy : {F : Code} → {m m′ m′′ : ℕ} →
        (f : Fin m′′ → μ F m′) → (t : μ F m′) →
        (y : Fin m) →
        (m≤m′′ : m ≤ m′′) →
        f (inject≤ y m≤m′′) ≡
        f-for f t (inject≤ (fromℕ m) (s≤s m≤m′′))
              (inject≤ (inject≤ y (n≤m+n 1 m)) (s≤s m≤m′′))
f-for-xy {m = m} f t y m≤m′′
  with thick₂ (fromℕ m) (inject≤ y (n≤m+n 1 m))
     | inspect (thick₂ (fromℕ m)) (inject≤ y (n≤m+n 1 m))
f-for-xy {m = m} f t y m≤m′′ | inj₁ m≡y | [ _ ]
  with cong toℕ m≡y | bounded y
... | eq | y<m rewrite inject≤-lemma y (n≤m+n 1 m) | to-from m
  = ⊥-elim (bot-f y<m eq)
f-for-xy f t y m≤m′′ | inj₂ (y′ , thinmy′≡y) | [ thick₂my≡inj₂y′ ]
  with inject-thick′ m≤m′′ thinmy′≡y thick₂my≡inj₂y′ 
... | eq′ rewrite eq′ | thinmy′≡y→y′≡y thinmy′≡y = refl

f-for-xy′ : {F : Code} → {m m′ m′′ : ℕ} →
        (f : Fin m′′ → μ F m′) → (t : μ F m′) →
        (y : Fin m) →
        (m≤m′′ : m ≤ m′′) →
        f (inject≤ y m≤m′′) ≡
        f-for f t (inject≤ (fromℕ m) (s≤s m≤m′′))
              (inject≤ y (≤-step m≤m′′))
f-for-xy′ {m = m} f t y m≤m′′ =
  begin
    f (inject≤ y m≤m′′)
  ≡⟨ f-for-xy f t y m≤m′′ ⟩
    f-for f t (inject≤ (fromℕ m) (s≤s m≤m′′))
          (inject≤ (inject≤ y (n≤m+n 1 m)) (s≤s m≤m′′))
  ≡⟨ cong (f-for f t (inject≤ (fromℕ m) (s≤s m≤m′′)))
      (inject≤-trans y (n≤m+n 1 m) (s≤s m≤m′′) (≤-step m≤m′′)) ⟩
    f-for f t (inject≤ (fromℕ m) (s≤s m≤m′′))
          (inject≤ y (≤-step m≤m′′))
  ∎

private
  mg-Unit : {m n : ℕ} → {Γ : Cxt {m} n} →
          mgt {m′ = m} anil Γ Unit TUnit
  mg-Unit m₀′′ m₀′ m≤m₀′′ f₀ .TUnit .Unit Unit =
    (m₀′′ , m₀′′ , m≤m₀′′ , m≤m₀′′ , Refl m≤m₀′′ , m≤m₀′′ , f₀ ,
     (λ x → sym (append-nil′ m≤m₀′′ f₀ (inject≤ x m≤m₀′′))) , refl)

  mg-Var : {m n : ℕ} → {Γ : Cxt {m} n} → (x : Fin n) →
        mgt anil Γ (Var x) (lookup x Γ)
  mg-Var {Γ = Γ} x m₀′′ m₀′ m≤m₀′′ f₀ ._ .(Var x) (Var .x) =
    (m₀′′ , m₀′′ , m≤m₀′′ , m≤m₀′′ , Refl m≤m₀′′ , m≤m₀′′ , f₀ ,
     (λ x → sym (append-nil′ m≤m₀′′ f₀ (inject≤ x m≤m₀′′))) ,
     sym (lookupFun≤CxtCommute x Γ f₀ m≤m₀′′))

  mg-Lam : {m n m′ m′′ : ℕ} → {Γ : Cxt {m} n} →
        (s₁ : WellScoped (suc n)) → (t₁ : Type m′) →
        (m≤m′′ : m ≤ m′′) →
        (σ : AList TypeF (suc m′′) m′) →
        (w₁ : WellTyped (applySubCxt≤ σ (s≤s m≤m′′) (⟪ fromℕ m ⟫ ∷ liftCxt 1 Γ)) t₁) →
        (er₁ : erase w₁ s₁) →
        (mgσ₁ : mgt σ (⟪ fromℕ m ⟫ ∷ liftCxt 1 Γ) s₁ t₁) →
        mgt σ Γ (Lam s₁) (applySub≤ σ (s≤s m≤m′′) ⟪ fromℕ m ⟫ ⇒ t₁)
  mg-Lam {m} {n} {m′} {m′′} {Γ} s₁ t₁ m≤m′′ σ w₁ er₁ mgσ₁ m₀′′ m₀′ m≤m₀′′ f₀
         ._ ._ (Lam {t₁ = t₁′} {t₂ = t₂′} w₁′ .s₁ er₀) =
    let a = mgσ₁ (suc m₀′′) m₀′ (s≤s m≤m₀′′) f₀′ t₁′ W₁ er′
    in g a
    where 1+m≤1+m₀′′ : suc m ≤ suc m₀′′
          1+m≤1+m₀′′ = s≤s m≤m₀′′
          x : Fin (suc m₀′′)
          x = inject≤ (fromℕ m) 1+m≤1+m₀′′
          f₀′ : Fin (suc m₀′′) → Type m₀′
          f₀′ = f-for f₀ t₂′ x
          W₁ : WellTyped
                (applyFunCxt≤ f₀′ (s≤s m≤m₀′′) (⟪ fromℕ m ⟫ ∷ liftCxt 1 Γ))
                t₁′
          W₁ rewrite f-for-xx f₀ t₂′ x | skip-fresh Γ m≤m₀′′ f₀ t₂′ = w₁′
          er′ : erase W₁ s₁
          er′ rewrite f-for-xx f₀ t₂′ x | skip-fresh Γ m≤m₀′′ f₀ t₂′ = er₀
          g : (Σ[ k′′ ∈ ℕ ]
              Σ[ k′′′ ∈ ℕ ]
              Σ[ m′≤k′′ ∈ m′ ≤ k′′ ]
              Σ[ 1+m′′≤k′′′ ∈ suc m′′ ≤ k′′′ ]
              Σ[ para ∈ 1+m′′≤k′′′ // m′≤k′′ ]
              Σ[ 1+m≤k′′′ ∈ suc m ≤ k′′′ ]
              Σ[ f′ ∈ (Fin k′′ → Type m₀′) ] 
              (((x : Fin (suc m)) →
                f₀′ (inject≤ x (s≤s m≤m₀′′)) ≡
                (f′ +⟨ para ⟩′ (sub σ)) (inject≤ x 1+m≤k′′′)) ×
               t₁′ ≡ applyUnifier≤ f′ m′≤k′′ t₁)) →
              (Σ[ k′′ ∈ ℕ ]
              Σ[ k′′′ ∈ ℕ ]
              Σ[ m′≤k′′ ∈ m′ ≤ k′′ ]
              Σ[ 1+m′′≤k′′′ ∈ suc m′′ ≤ k′′′ ]
              Σ[ para ∈ 1+m′′≤k′′′ // m′≤k′′ ]
              Σ[ m≤k′′′ ∈ m ≤ k′′′ ]
              Σ[ f′ ∈ (Fin k′′ → Type m₀′) ] 
              (((x : Fin m) →
                f₀ (inject≤ x m≤m₀′′) ≡
                (f′ +⟨ para ⟩′ (sub σ)) (inject≤ x m≤k′′′)) ×
               t₂′ ⇒ t₁′ ≡ applyUnifier≤ f′ m′≤k′′
                     (applySub≤ σ (s≤s m≤m′′) ⟪ fromℕ m ⟫ ⇒ t₁)))
          g (k′′ , k′′′ , m′≤k′′ , 1+m′′≤k′′′ , para₁ , 1+m≤k′′′ , f₁′ , eq-f₁′ , eq₁)
            = (k′′ , k′′′ , m′≤k′′ , 1+m′′≤k′′′ , para₁ , m≤k′′′ , f₁′ ,
               eq-y , eq-t)
            where m≤k′′′ : m ≤ k′′′
                  m≤k′′′ = ≤-trans m≤m′′ (≤-trans (n≤m+n 1 m′′) 1+m′′≤k′′′)
                  eq-y : (y : Fin m) →
                       f₀ (inject≤ y m≤m₀′′) ≡
                       (f₁′ +⟨ para₁ ⟩′ sub σ) (inject≤ y m≤k′′′)
                  eq-y y =
                    begin
                      f₀ (inject≤ y m≤m₀′′)
                    ≡⟨ f-for-xy f₀ t₂′ y m≤m₀′′ ⟩
                      f₀′ (inject≤ (inject≤ y (n≤m+n 1 m)) (s≤s m≤m₀′′))
                    ≡⟨ eq-f₁′ (inject≤ y (n≤m+n 1 m)) ⟩
                      (f₁′ +⟨ para₁ ⟩′ sub σ)
                        (inject≤ (inject≤ y (n≤m+n 1 m)) 1+m≤k′′′)
                    ≡⟨ cong (f₁′ +⟨ para₁ ⟩′ sub σ)
                           (inject≤-trans y (n≤m+n 1 m) 1+m≤k′′′ m≤k′′′) ⟩
                      (f₁′ +⟨ para₁ ⟩′ sub σ) (inject≤ y m≤k′′′)
                    ∎
                  eq-t₂ : t₂′ ≡
                          applyUnifier≤ f₁′ m′≤k′′
                           (applySub≤ σ (s≤s m≤m′′) ⟪ fromℕ m ⟫)
                  eq-t₂ =
                    begin
                      t₂′
                    ≡⟨ sym (f-for-xx f₀ t₂′ (inject≤ (fromℕ m) (s≤s m≤m₀′′))) ⟩
                      f₀′ (inject≤ (fromℕ m) (s≤s m≤m₀′′))
                    ≡⟨ refl ⟩
                      applyUnifier≤ f₀′ (s≤s m≤m₀′′) ⟪ fromℕ m ⟫
                    ≡⟨ fuse-applyUnifier′≤ f₀′ f₁′ σ 1+m′′≤k′′′ m′≤k′′ para₁ (s≤s m≤m₀′′) (s≤s m≤m′′) 1+m≤k′′′ eq-f₁′ ⟪ fromℕ m ⟫ ⟩
                      applyUnifier≤ f₁′ m′≤k′′
                        (applySub≤ σ (s≤s m≤m′′) ⟪ fromℕ m ⟫)
                    ∎
                  eq-t : t₂′ ⇒ t₁′ ≡
                         applyUnifier≤ f₁′ m′≤k′′
                          (applySub≤ σ (s≤s m≤m′′) ⟪ fromℕ m ⟫ ⇒ t₁)
                  eq-t =
                    begin
                      t₂′ ⇒ t₁′
                    ≡⟨ cong₂ _⇒_ eq-t₂ eq₁ ⟩
                      applyUnifier≤ f₁′ m′≤k′′
                       (applySub≤ σ (s≤s m≤m′′) ⟪ fromℕ m ⟫) ⇒
                      applyUnifier≤ f₁′ m′≤k′′ t₁
                    ≡⟨ refl ⟩
                      applyUnifier≤ f₁′ m′≤k′′
                       (applySub≤ σ (s≤s m≤m′′) ⟪ fromℕ m ⟫ ⇒ t₁)
                    ∎

private
  mg-App : {m n m₁′ m₂′ m₁′′ m₂′′ m₃′ : ℕ} → {Γ : Cxt {m} n} →
        (s₁ : WellScoped n) →
        (m≤m₁′′ : m ≤ m₁′′) →
        (σ₁ : AList TypeF m₁′′ m₁′) →
        (t₁ : Type m₁′) →
        (w₁ : WellTyped (applySubCxt≤ σ₁ m≤m₁′′ Γ) t₁) →
        (er₁ : erase w₁ s₁) →
        (mgσ₁ : mgt σ₁ Γ s₁ t₁) →

        (s₂ : WellScoped n) →
        (m₁′≤m₂′′ : m₁′ ≤ m₂′′) →
        (σ₂ : AList TypeF m₂′′ m₂′) →
        (t₂ : Type m₂′) →
        (w₂ : WellTyped (applySubCxt≤ σ₂ m₁′≤m₂′′
                         (applySubCxt≤ σ₁ m≤m₁′′ Γ)) t₂) →
        (er₂ : erase w₂ s₂) →
        (mgσ₂ : mgt σ₂ (applySubCxt≤ σ₁ m≤m₁′′ Γ) s₂ t₂) →

        (σ₃ : AList TypeF (suc m₂′) m₃′) →
        let m₂′≤1+m₂′ = n≤m+n 1 m₂′ in
        (mg (lift≤ m₂′≤1+m₂′ (applySub≤ σ₂ m₁′≤m₂′′ t₁))
            (lift≤ m₂′≤1+m₂′ t₂ ⇒ ⟪ fromℕ m₂′ ⟫)
            (sub σ₃)) →
        let σ₂₁ = σ₂ +⟨ proj₂ (proj₂ (σ→// σ₁ m₁′≤m₂′′)) ⟩ σ₁ in
        mgt (σ₃ +⟨ proj₂ (proj₂ (σ→// σ₂₁ m₂′≤1+m₂′)) ⟩ σ₂₁ )
          Γ (App s₁ s₂) (applySub σ₃ ⟪ fromℕ m₂′ ⟫)
  mg-App {m} {m₂′ = m₂′} {Γ = Γ} s₁ m≤m₁′′ σ₁ t₁ w₁ er₁ mgσ₁
         s₂ m₁′≤m₂′′ σ₂ t₂ w₂ er₂ mgσ₂
         σ₃ (σ₃σ₂t₁≡σ₃t₂⇒σ₃t , mg-unify)
         l′′′ l′ m≤l′′′ f₀ t′ .(App w₁′ w₂′)
         (App {t₂ = t₂′} w₁′ .s₁ w₂′ .s₂ er₁′ er₂′)
    with mgσ₁ l′′′ l′ m≤l′′′ f₀ (t₂′ ⇒ t′) w₁′ er₁′
  ... | (k₁′′ , k₁′′′ , m₁′≤k₁′′ , m₁′′≤k₁′′′ , para₁ , m≤k₁′′′ , f₁′ , eq-f₁′ , eq₁)
    rewrite funCxtTrans′ f₀ Γ f₁′ σ₁ m₁′′≤k₁′′′ m₁′≤k₁′′ para₁ m≤l′′′ m≤m₁′′ m≤k₁′′′ eq-f₁′
    with mgσ₂ k₁′′ l′ m₁′≤k₁′′ f₁′ t₂′ w₂′ er₂′
  ... | (k₂′′ , k₂′′′ , m₂′≤k₂′′ , m₂′′≤k₂′′′ , para₂ , m₁′≤k₂′′′ , f₂′ , eq-f₂′ , eq₂)
    with mg-unify l′ (suc k₂′′)
                  (f-for f₂′ t′ (inject≤ (fromℕ m₂′) (s≤s m₂′≤k₂′′)))
                  (s≤s m₂′≤k₂′′)
                  eq
    where f₂′′ : (x : Fin (suc k₂′′)) → Type l′
          f₂′′ = f-for f₂′ t′ (inject≤ (fromℕ m₂′) (s≤s m₂′≤k₂′′))
          eq : applyUnifier≤ f₂′′ (s≤s m₂′≤k₂′′)
                 (lift≤ (n≤m+n 1 m₂′)
                   (applySub≤ σ₂ m₁′≤m₂′′ t₁)) ≡
               applyUnifier≤ f₂′′ (s≤s m₂′≤k₂′′)
                 (lift≤ (n≤m+n 1 m₂′) t₂) ⇒
               applyUnifier≤ f₂′′ (s≤s m₂′≤k₂′′) ⟪ fromℕ m₂′ ⟫
          eq =
           begin
             applyUnifier≤ f₂′′ (s≤s m₂′≤k₂′′)
               (lift≤ (n≤m+n 1 m₂′)
                 (applySub≤ σ₂ m₁′≤m₂′′ t₁))
           ≡⟨ cong (applyUnifier f₂′′)
               (fuse-lift (n≤m+n 1 m₂′) (s≤s m₂′≤k₂′′) (≤-step m₂′≤k₂′′)
                (applySub≤ σ₂ m₁′≤m₂′′ t₁)) ⟩
             applyUnifier≤ f₂′′ (≤-step m₂′≤k₂′′)
              (applySub≤ σ₂ m₁′≤m₂′′ t₁)
           ≡⟨ cata-ext′′ f₂′′ f₂′ (≤-step m₂′≤k₂′′) m₂′≤k₂′′
               (λ y → sym (f-for-xy′ f₂′ t′ y m₂′≤k₂′′))
               (applySub≤ σ₂ m₁′≤m₂′′ t₁) ⟩
             applyUnifier≤ f₂′ m₂′≤k₂′′
              (applySub≤ σ₂ m₁′≤m₂′′ t₁)
           ≡⟨ cong (applyUnifier f₂′)
               (liftSubCommute σ₂ para₂ (lift≤ m₁′≤m₂′′ t₁)) ⟩
             applyUnifier f₂′
              (applySub (liftAList≤ para₂ σ₂)
                (lift≤ m₂′′≤k₂′′′ (lift≤ m₁′≤m₂′′ t₁)))
           ≡⟨ cong (λ t → applyUnifier f₂′
                             (applySub (liftAList≤ para₂ σ₂) t))
               (fuse-lift m₁′≤m₂′′ m₂′′≤k₂′′′ m₁′≤k₂′′′ t₁) ⟩
             applyUnifier f₂′
              (applySub (liftAList≤ para₂ σ₂)
                (lift≤ m₁′≤k₂′′′ t₁))
           ≡⟨ sym (applyUnifierAppend′ f₂′ σ₂ para₂ (lift≤ m₁′≤k₂′′′ t₁)) ⟩
             applyUnifier (f₂′ +⟨ para₂ ⟩′ (sub σ₂))
              (lift≤ m₁′≤k₂′′′ t₁)
           ≡⟨ sym (cata-ext′′ f₁′ (f₂′ +⟨ para₂ ⟩′ (sub σ₂))
                             m₁′≤k₁′′ m₁′≤k₂′′′ eq-f₂′ t₁) ⟩
             applyUnifier≤ f₁′ m₁′≤k₁′′ t₁
           ≡⟨ sym eq₁ ⟩
             t₂′ ⇒ t′
           ≡⟨ cong₂ _⇒_ eq₂
              (sym (f-for-xx f₂′ t′ (inject≤ (fromℕ m₂′) (s≤s m₂′≤k₂′′)))) ⟩
             applyUnifier≤ f₂′ m₂′≤k₂′′ t₂ ⇒
             applyUnifier≤ f₂′′ (s≤s m₂′≤k₂′′) ⟪ fromℕ m₂′ ⟫
           ≡⟨ cong (λ t → t ⇒ applyUnifier≤ f₂′′ (s≤s m₂′≤k₂′′) ⟪ fromℕ m₂′ ⟫)
               (cata-ext′′ f₂′ f₂′′ m₂′≤k₂′′ (≤-step m₂′≤k₂′′) (λ y → f-for-xy′ f₂′ t′ y m₂′≤k₂′′) t₂) ⟩
             applyUnifier≤ f₂′′ (≤-step m₂′≤k₂′′) t₂ ⇒
             applyUnifier≤ f₂′′ (s≤s m₂′≤k₂′′) ⟪ fromℕ m₂′ ⟫
           ≡⟨ cong (λ t → applyUnifier f₂′′ t ⇒ applyUnifier≤ f₂′′ (s≤s m₂′≤k₂′′) ⟪ fromℕ m₂′ ⟫)
               (sym (fuse-lift (n≤m+n 1 m₂′) (s≤s m₂′≤k₂′′) (≤-step m₂′≤k₂′′) t₂)) ⟩
             applyUnifier≤ f₂′′ (s≤s m₂′≤k₂′′)
               (lift≤ (n≤m+n 1 m₂′) t₂) ⇒
             applyUnifier≤ f₂′′ (s≤s m₂′≤k₂′′) ⟪ fromℕ m₂′ ⟫
           ∎
  ... | (k₃′′ , m₃′≤k₃′′ , para₃ , f₃′ , eq-f₃′) =
    let tuple₁ = σ→// σ₁ m₁′≤m₂′′
        m₂′′′ = proj₁ tuple₁
        m₁′′≤m₂′′′ = proj₁ (proj₂ tuple₁)
        para₁′ = proj₂ (proj₂ tuple₁)
        σ₂₁ = σ₂ +⟨ para₁′ ⟩ σ₁
        m₂′≤1+m₂′ = n≤m+n 1 m₂′
        tuple₂ = σ→// σ₂₁ m₂′≤1+m₂′
        1+m₂′′′ = proj₁ tuple₂
        m₂′′′≤1+m₂′′′ = proj₁ (proj₂ tuple₂)
        para₂′ = proj₂ (proj₂ tuple₂)
        σ₃₂₁ = σ₃ +⟨ para₂′ ⟩ σ₂₁
        tuple₃ = σ→// σ₃₂₁ m₃′≤k₃′′ 
        m₄′ = proj₁ tuple₃
        1+m₂′′′≤m₄′ = proj₁ (proj₂ tuple₃)
        para₃′ = proj₂ (proj₂ tuple₃)
        m≤m₂′′′ = ≤-trans m≤m₁′′ m₁′′≤m₂′′′
        m≤1+m₂′′′ = ≤-trans m≤m₂′′′ m₂′′′≤1+m₂′′′
        m≤m₄′ = ≤-trans m≤1+m₂′′′  1+m₂′′′≤m₄′
        f₂′′ = f-for f₂′ t′ (inject≤ (fromℕ m₂′) (s≤s m₂′≤k₂′′))
    in (k₃′′ , m₄′ , m₃′≤k₃′′ , 1+m₂′′′≤m₄′ , para₃′ , m≤m₄′ , f₃′ ,
        (λ x →
         begin
           f₀ (inject≤ x m≤l′′′)
         ≡⟨ eq-f₁′ x ⟩
           applyUnifier (f₁′ +⟨ para₁ ⟩′ sub σ₁)
             (lift≤ m≤k₁′′′ ⟪ x ⟫)
         ≡⟨ applyUnifierAppend′ f₁′ σ₁ para₁ (lift≤ m≤k₁′′′ ⟪ x ⟫) ⟩
           applyUnifier f₁′
            (applySub (liftAList≤ para₁ σ₁)
             (lift≤ m≤k₁′′′ ⟪ x ⟫))
         ≡⟨ cong (λ t →
                  applyUnifier f₁′
                  (applySub (liftAList≤ para₁ σ₁) t))
               (sym (fuse-lift m≤m₁′′ m₁′′≤k₁′′′ m≤k₁′′′ ⟪ x ⟫)) ⟩
           applyUnifier f₁′
            (applySub (liftAList≤ para₁ σ₁)
             (lift≤ m₁′′≤k₁′′′ (lift≤ m≤m₁′′ ⟪ x ⟫)))
         ≡⟨ cong (applyUnifier f₁′)
             (sym (liftSubCommute σ₁ para₁ (lift≤ m≤m₁′′ ⟪ x ⟫))) ⟩
           applyUnifier f₁′
            (lift≤ m₁′≤k₁′′ (applySub σ₁ (lift≤ m≤m₁′′ ⟪ x ⟫)))
         ≡⟨ cata-ext′′ f₁′ (f₂′ +⟨ para₂ ⟩′ sub σ₂) m₁′≤k₁′′ m₁′≤k₂′′′
                      eq-f₂′ (applySub σ₁ (lift≤ m≤m₁′′ ⟪ x ⟫)) ⟩
           applyUnifier (f₂′ +⟨ para₂ ⟩′ sub σ₂)
            (lift≤ m₁′≤k₂′′′ (applySub σ₁ (lift≤ m≤m₁′′ ⟪ x ⟫)))
         ≡⟨ applyUnifierAppend′ f₂′ σ₂ para₂
              (lift≤ m₁′≤k₂′′′ (applySub σ₁ (lift≤ m≤m₁′′ ⟪ x ⟫))) ⟩
           applyUnifier f₂′ (applySub (liftAList≤ para₂ σ₂)
            (lift≤ m₁′≤k₂′′′ (applySub σ₁ (lift≤ m≤m₁′′ ⟪ x ⟫))))
         ≡⟨ cong (λ t → applyUnifier f₂′
                           (applySub (liftAList≤ para₂ σ₂) t))
             (sym (fuse-lift m₁′≤m₂′′ m₂′′≤k₂′′′ m₁′≤k₂′′′
                             (applySub σ₁ (lift≤ m≤m₁′′ ⟪ x ⟫)))) ⟩
           applyUnifier f₂′ (applySub (liftAList≤ para₂ σ₂)
            (lift≤ m₂′′≤k₂′′′ (lift≤ m₁′≤m₂′′ (applySub σ₁ (lift≤ m≤m₁′′ ⟪ x ⟫)))))
         ≡⟨ cong (applyUnifier f₂′) (sym (liftSubCommute σ₂ para₂ (lift≤ m₁′≤m₂′′ (applySub σ₁ (lift≤ m≤m₁′′ ⟪ x ⟫))))) ⟩
           applyUnifier f₂′ (lift≤ m₂′≤k₂′′
            (applySub σ₂ (lift≤ m₁′≤m₂′′ (applySub σ₁ (lift≤ m≤m₁′′ ⟪ x ⟫)))))
         ≡⟨ cata-ext′′ f₂′ f₂′′ m₂′≤k₂′′ (≤-step m₂′≤k₂′′)
             (λ y → f-for-xy′ f₂′ t′ y m₂′≤k₂′′)
             (applySub σ₂ (lift≤ m₁′≤m₂′′ (applySub σ₁ (lift≤ m≤m₁′′ ⟪ x ⟫)))) ⟩
           applyUnifier f₂′′ (lift≤ (≤-step m₂′≤k₂′′)
            (applySub σ₂ (lift≤ m₁′≤m₂′′ (applySub σ₁ (lift≤ m≤m₁′′ ⟪ x ⟫)))))
         ≡⟨ cong (applyUnifier f₂′′) (sym (fuse-lift m₂′≤1+m₂′ (s≤s m₂′≤k₂′′) (≤-step m₂′≤k₂′′) (applySub σ₂ (lift≤ m₁′≤m₂′′ (applySub σ₁ (lift≤ m≤m₁′′ ⟪ x ⟫)))))) ⟩
           applyUnifier≤ f₂′′ (s≤s m₂′≤k₂′′) (lift≤ m₂′≤1+m₂′
            (applySub σ₂ (lift≤ m₁′≤m₂′′ (applySub σ₁ (lift≤ m≤m₁′′ ⟪ x ⟫)))))
         ≡⟨ cong (λ t →
                  applyUnifier≤ f₂′′ (s≤s m₂′≤k₂′′)
                  (lift≤ m₂′≤1+m₂′ (applySub σ₂ t)))
               (liftSubCommute σ₁ para₁′ (lift≤ m≤m₁′′ ⟪ x ⟫)) ⟩
           applyUnifier≤ f₂′′ (s≤s m₂′≤k₂′′) (lift≤ m₂′≤1+m₂′
            (applySub σ₂ (applySub (liftAList≤ para₁′ σ₁)
             (lift≤ m₁′′≤m₂′′′ (lift≤ m≤m₁′′ ⟪ x ⟫)))))
         ≡⟨ cong (λ t →
                  applyUnifier≤ f₂′′ (s≤s m₂′≤k₂′′) (lift≤ m₂′≤1+m₂′
                   (applySub σ₂
                    (applySub (liftAList≤ para₁′ σ₁) t))))
               (fuse-lift m≤m₁′′ m₁′′≤m₂′′′ m≤m₂′′′ ⟪ x ⟫) ⟩
           applyUnifier≤ f₂′′ (s≤s m₂′≤k₂′′) (lift≤ m₂′≤1+m₂′
            (applySub σ₂ (applySub (liftAList≤ para₁′ σ₁)
             (lift≤ m≤m₂′′′ ⟪ x ⟫))))
         ≡⟨ cong (λ t → applyUnifier≤ f₂′′ (s≤s m₂′≤k₂′′) (lift≤ m₂′≤1+m₂′ t))
             (sub-++-commute′ (lift≤ m≤m₂′′′ ⟪ x ⟫) σ₁ m₁′≤m₂′′ m₁′′≤m₂′′′ para₁′ σ₂) ⟩
           applyUnifier≤ f₂′′ (s≤s m₂′≤k₂′′) (lift≤ m₂′≤1+m₂′
            (applySub (σ₂ +⟨ para₁′ ⟩ σ₁) (lift≤ m≤m₂′′′ ⟪ x ⟫)))
         ≡⟨ cata-ext′′ f₂′′ (f₃′ +⟨ para₃ ⟩′ sub σ₃)
             (s≤s m₂′≤k₂′′) (s≤s m₂′≤k₂′′) eq-f₃′
             (lift≤ m₂′≤1+m₂′ (applySub σ₂₁ (lift≤ m≤m₂′′′ ⟪ x ⟫))) ⟩
           applyUnifier≤ (f₃′ +⟨ para₃ ⟩′ sub σ₃) (s≤s m₂′≤k₂′′)
            (lift≤ m₂′≤1+m₂′ (applySub σ₂₁ (lift≤ m≤m₂′′′ ⟪ x ⟫)))
         ≡⟨ applyUnifierAppend′ f₃′ σ₃ para₃
             (lift≤ (s≤s m₂′≤k₂′′)
              (lift≤ m₂′≤1+m₂′ (applySub σ₂₁ (lift≤ m≤m₂′′′ ⟪ x ⟫)))) ⟩
           applyUnifier f₃′ (applySub (liftAList≤ para₃ σ₃)
            (lift≤ (s≤s m₂′≤k₂′′)
             (lift≤ m₂′≤1+m₂′ (applySub σ₂₁ (lift≤ m≤m₂′′′ ⟪ x ⟫)))))
         ≡⟨ cong (λ t →
                  applyUnifier f₃′
                  (applySub (liftAList≤ para₃ σ₃) (lift≤ (s≤s m₂′≤k₂′′) t)))
             (liftSubCommute σ₂₁ para₂′ (lift≤ m≤m₂′′′ ⟪ x ⟫)) ⟩
           applyUnifier f₃′ (applySub (liftAList≤ para₃ σ₃)
            (lift≤ (s≤s m₂′≤k₂′′)
             (applySub (liftAList≤ para₂′ σ₂₁)
              (lift≤ m₂′′′≤1+m₂′′′ (lift≤ m≤m₂′′′ ⟪ x ⟫)))))
         ≡⟨ cong (λ t →
                  applyUnifier f₃′
                  (applySub (liftAList≤ para₃ σ₃)
                   (lift≤ (s≤s m₂′≤k₂′′) (applySub (liftAList≤ para₂′ σ₂₁) t))))
               (fuse-lift m≤m₂′′′ m₂′′′≤1+m₂′′′ m≤1+m₂′′′ ⟪ x ⟫) ⟩
           applyUnifier f₃′ (applySub (liftAList≤ para₃ σ₃)
            (lift≤ (s≤s m₂′≤k₂′′)
             (applySub (liftAList≤ para₂′ σ₂₁)
              (lift≤ m≤1+m₂′′′ ⟪ x ⟫))))
         ≡⟨ cong (applyUnifier f₃′)
             (sym (liftSubCommute σ₃ para₃
                   (applySub (liftAList≤ para₂′ σ₂₁)
                    (lift≤ m≤1+m₂′′′ ⟪ x ⟫)))) ⟩
           applyUnifier≤ f₃′ m₃′≤k₃′′
            (applySub σ₃ (applySub (liftAList≤ para₂′ σ₂₁)
             (lift≤ m≤1+m₂′′′ ⟪ x ⟫)))
         ≡⟨ cong (applyUnifier≤ f₃′ m₃′≤k₃′′)
             (sub-++-commute′ (lift≤ m≤1+m₂′′′ ⟪ x ⟫) σ₂₁ m₂′≤1+m₂′ m₂′′′≤1+m₂′′′
                             para₂′ σ₃) ⟩
           applyUnifier≤ f₃′ m₃′≤k₃′′
            (applySub (σ₃ +⟨ para₂′ ⟩ σ₂₁)
             (lift≤ m≤1+m₂′′′ ⟪ x ⟫))
         ≡⟨ cong (applyUnifier f₃′)
             (liftSubCommute σ₃₂₁ para₃′ (lift≤ m≤1+m₂′′′ ⟪ x ⟫)) ⟩
           applyUnifier f₃′ (applySub (liftAList≤ para₃′ σ₃₂₁)
             (lift≤ 1+m₂′′′≤m₄′ (lift≤ m≤1+m₂′′′ ⟪ x ⟫)))
         ≡⟨ cong (λ t →
                  applyUnifier f₃′
                  (applySub (liftAList≤ para₃′ σ₃₂₁) t))
               (fuse-lift m≤1+m₂′′′ 1+m₂′′′≤m₄′ m≤m₄′ ⟪ x ⟫) ⟩

           applyUnifier f₃′ (applySub (liftAList≤ para₃′ σ₃₂₁)
             (lift≤ m≤m₄′ ⟪ x ⟫))
         ≡⟨ sym (applyUnifierAppend′ f₃′ σ₃₂₁ para₃′ ⟪ inject≤ x m≤m₄′ ⟫) ⟩
           (f₃′ +⟨ para₃′ ⟩′ (sub σ₃₂₁)) (inject≤ x m≤m₄′)
         ∎) ,
        (begin
           t′
         ≡⟨ sym (f-for-xx f₂′ t′ (inject≤ (fromℕ m₂′) (s≤s m₂′≤k₂′′))) ⟩
           f₂′′ (inject≤ (fromℕ m₂′) (s≤s m₂′≤k₂′′))
         ≡⟨ eq-f₃′ (fromℕ m₂′) ⟩
           applyUnifier (f₃′ +⟨ para₃ ⟩′ sub σ₃)
             (lift≤ (s≤s m₂′≤k₂′′) ⟪ fromℕ m₂′ ⟫)
         ≡⟨ applyUnifierAppend′ f₃′ σ₃ para₃ (lift≤ (s≤s m₂′≤k₂′′) ⟪ fromℕ m₂′ ⟫) ⟩
           applyUnifier f₃′ (applySub (liftAList≤ para₃ σ₃)
             (lift≤ (s≤s m₂′≤k₂′′) ⟪ fromℕ m₂′ ⟫))
         ≡⟨ cong (applyUnifier f₃′) (sym (liftSubCommute σ₃ para₃ ⟪ fromℕ m₂′ ⟫)) ⟩
           applyUnifier≤ f₃′ m₃′≤k₃′′ (applySub σ₃ ⟪ fromℕ m₂′ ⟫)
         ∎))

private
  illtyped-Lam : {m n : ℕ} → (Γ : Cxt {m} n) →
          (s₁ : WellScoped (1 + n)) →
          (ill-s₁ : untypable (⟪ fromℕ m ⟫ ∷ liftCxt 1 Γ) s₁) →
          untypable Γ (Lam s₁)
  illtyped-Lam {m} Γ s₁ ill-s₁ m′′ m′ m≤m′′ f ._ ._
                   (Lam {t₁ = t₁} {t₂ = t₂} w₁ .s₁ er)
    = ill-s₁ (suc m′′) m′ 1+m≤1+m′′ f′ t₁ W₁ er′
        where 1+m≤1+m′′ : suc m ≤ suc m′′
              1+m≤1+m′′ = s≤s m≤m′′
              x : Fin (suc m′′)
              x = inject≤ (fromℕ m) 1+m≤1+m′′
              f′ : Fin (suc m′′) → Type m′
              f′ = f-for f t₂ x
              W₁ : WellTyped
                    (applyFunCxt≤ f′ (s≤s m≤m′′) (⟪ fromℕ m ⟫ ∷ liftCxt 1 Γ))
                    t₁
              W₁ rewrite f-for-xx f t₂ x | skip-fresh Γ m≤m′′ f t₂ = w₁
              er′ : erase W₁ s₁
              er′ rewrite f-for-xx f t₂ x | skip-fresh Γ m≤m′′ f t₂ = er

private
  illtyped-App₁ : {m n : ℕ} → (Γ : Cxt {m} n) →
          (s₁ s₂ : WellScoped n) →
          (ill-s₁ : untypable Γ s₁) →
          untypable Γ (App s₁ s₂)
  illtyped-App₁ Γ s₁ s₂ ill-s₁ m′′ m′ m≤m′′ f t₁
                .(App w₁ w₂) (App {t₂ = t₂} w₁ .s₁ w₂ .s₂ er₁ er₂) =
    ill-s₁ m′′ m′ m≤m′′ f (t₂ ⇒ t₁) w₁ er₁

private
  illtyped-App₂ : {m n m₁′ m₁′′ : ℕ} → (Γ : Cxt {m} n) →
          (s₁ s₂ : WellScoped n) →
          (t₁ : Type m₁′) →
          (m≤m₁′′ : m ≤ m₁′′) →
          (σ₁ : AList TypeF m₁′′ m₁′) →
          (mgσ₁ : mgt σ₁ Γ s₁ t₁) →
          (ill-s₂ : untypable (applySubCxt≤ σ₁ m≤m₁′′ Γ) s₂) →
          untypable Γ (App s₁ s₂)
  illtyped-App₂ Γ s₁ s₂ t₁ m≤m₁′′ σ₁ mgσ₁ ill-s₂ m′′ m′ m≤m′′ f t₁′
    .(App w₁ w₂) (App {t₂ = t₂} w₁ .s₁ w₂ .s₂ er₁ er₂)
    with mgσ₁ m′′ m′ m≤m′′ f (t₂ ⇒ t₁′) w₁ er₁
  ... | (k′′ , k′′′ , m₁′≤k′′ , m₁′′≤k′′′ , para₁ , m′′≤k′′′ , f₁′ , eq-f₁ , eq₁)
    = ill-s₂ k′′ m′ m₁′≤k′′ f₁′ t₂ W₂ er₂′
        where a : applyFunCxt≤ f m≤m′′ Γ ≡
                  applyFunCxt≤ f₁′ m₁′≤k′′ (applySubCxt≤ σ₁ m≤m₁′′ Γ)
              a = funCxtTrans′ f Γ f₁′ σ₁ m₁′′≤k′′′ m₁′≤k′′ para₁ m≤m′′ m≤m₁′′ m′′≤k′′′ eq-f₁
              W₂ : WellTyped
                    (applyFunCxt≤ f₁′ m₁′≤k′′ (applySubCxt≤ σ₁ m≤m₁′′ Γ))
                    t₂
              W₂ rewrite sym a = w₂
              er₂′ : erase W₂ s₂
              er₂′ rewrite sym a = er₂

  illtyped-unify : {m n m₁′ m₁′′ m₂′ m₂′′ : ℕ} → (Γ : Cxt {m} n) →
          (s₁ s₂ : WellScoped n) →
          (t₁ : Type m₁′) →
          (m≤m₁′′ : m ≤ m₁′′) →
          (m₁′≤m₂′′ : m₁′ ≤ m₂′′) →
          -- (m₂′≤1+m₂′ : m₂′ ≤ suc m₂′) →
          (σ₁ : AList TypeF m₁′′ m₁′) →
          (mgσ₁ : mgt σ₁ Γ s₁ t₁) →
          (σ₂ : AList TypeF m₂′′ m₂′) →
          (t₂ : Type m₂′) →
          -- (t : Type (suc m₂′)) →
          (mgσ₂ : mgt σ₂ (applySubCxt≤ σ₁ m≤m₁′′ Γ) s₂ t₂) →
          (ill-unify : ununifiable (lift≤ (n≤m+n 1 m₂′) (applySub≤ σ₂ m₁′≤m₂′′ t₁))
                                   (lift≤ (n≤m+n 1 m₂′) t₂ ⇒ ⟪ fromℕ m₂′ ⟫)) →
          untypable Γ (App s₁ s₂)
  illtyped-unify {m₂′ = m₂′} Γ s₁ s₂ t₁ m≤m₁′′ m₁′≤m₂′′ σ₁ mgσ₁ σ₂ t₂ mgσ₂ ill-unify m′′ m′ m≤m′′ f t′ .(App w₁ w₂) (App {t₂ = t₂′} w₁ .s₁ w₂ .s₂ er₁ er₂)
    with mgσ₁ m′′ m′ m≤m′′ f (t₂′ ⇒ t′) w₁ er₁
  ... | (k₁′′ , k₁′′′ , m₁′≤k₁′′ , m₁′′≤k₁′′′ , para₁ , m≤k₁′′′ , f₁′ , eq-f₁′ , eq₁)
    rewrite funCxtTrans′ f Γ f₁′ σ₁ m₁′′≤k₁′′′ m₁′≤k₁′′ para₁ m≤m′′ m≤m₁′′ m≤k₁′′′ eq-f₁′
    with mgσ₂ k₁′′ m′ m₁′≤k₁′′ f₁′ t₂′ w₂ er₂
  ... | (k₂′′ , k₂′′′ , m₂′≤k₂′′ , m₁′′≤k₂′′′ , para₂ , m≤k₂′′′ , f₂′ , eq-f₂′ , eq₂)
    = ill-unify m′ (suc k₂′′) (s≤s m₂′≤k₂′′) f₂′′ eq
    where
      y : Fin (suc m₂′)
      y = fromℕ m₂′
      f₂′′ : (x : Fin (suc k₂′′)) → Type m′
      f₂′′ = f-for f₂′ t′ (inject≤ y (s≤s m₂′≤k₂′′))
      m₂′≤1+m₂′ = n≤m+n 1 m₂′
      eq : applyUnifier≤ f₂′′ (s≤s m₂′≤k₂′′)
             (lift≤ m₂′≤1+m₂′ (applySub≤ σ₂ m₁′≤m₂′′ t₁)) ≡
           applyUnifier≤ f₂′′ (s≤s m₂′≤k₂′′) (lift≤ m₂′≤1+m₂′ t₂ ⇒ ⟪ y ⟫)
      eq = begin
             applyUnifier≤ f₂′′ (s≤s m₂′≤k₂′′)
               (lift≤ m₂′≤1+m₂′ (applySub≤ σ₂ m₁′≤m₂′′ t₁))
           ≡⟨ skip-fresh-type m₂′≤k₂′′ f₂′ t′ (applySub≤ σ₂ m₁′≤m₂′′ t₁) ⟩
             applyUnifier≤ f₂′ m₂′≤k₂′′ (applySub≤ σ₂ m₁′≤m₂′′ t₁)
           ≡⟨ sym (fuse-applyUnifier′≤ f₁′ f₂′ σ₂ m₁′′≤k₂′′′ m₂′≤k₂′′ para₂ m₁′≤k₁′′ m₁′≤m₂′′ m≤k₂′′′ eq-f₂′ t₁) ⟩
             applyUnifier≤ f₁′ m₁′≤k₁′′ t₁
           ≡⟨ sym eq₁ ⟩
             t₂′ ⇒ t′
           ≡⟨ cong (λ t → t ⇒ t′) eq₂ ⟩
             applyUnifier≤ f₂′ m₂′≤k₂′′ t₂ ⇒ t′
           ≡⟨ cong (λ t → t ⇒ t′) (sym (skip-fresh-type m₂′≤k₂′′ f₂′ t′ t₂)) ⟩
             applyUnifier≤ f₂′′ (s≤s m₂′≤k₂′′) (lift≤ m₂′≤1+m₂′ t₂) ⇒ t′
           ≡⟨ cong
                 (_⇒_ (applyUnifier≤ f₂′′ (s≤s m₂′≤k₂′′)
                         (lift≤ m₂′≤1+m₂′ t₂)))
                 (sym (f-for-xx f₂′ t′ (inject≤ y (s≤s m₂′≤k₂′′)))) ⟩
             applyUnifier≤ f₂′′ (s≤s m₂′≤k₂′′) (lift≤ m₂′≤1+m₂′ t₂) ⇒
             applyUnifier≤ f₂′′ (s≤s m₂′≤k₂′′) ⟪ y ⟫
           ≡⟨ refl ⟩
             applyUnifier≤ f₂′′ (s≤s m₂′≤k₂′′) (lift≤ m₂′≤1+m₂′ t₂ ⇒ ⟪ y ⟫)
           ∎

mutual
  infer : (m : ℕ) → {n : ℕ} → (Γ : Cxt {m} n) → (s : WellScoped n) →
         untypable Γ s
       ⊎ Σ[ m′′ ∈ ℕ ] Σ[ m′ ∈ ℕ ] Σ[ m≤m′′ ∈ m ≤ m′′ ]
         Σ[ σ ∈ AList TypeF m′′ m′ ] Σ[ t ∈ Type m′ ]
         Σ[ w ∈ WellTyped (applySubCxt≤ σ m≤m′′ Γ) t ]
         (erase w s × mgt σ Γ s t)
  infer m Γ Unit = infer-Unit m Γ
  infer m Γ (Var x) = infer-Var m Γ x
  infer m Γ (Lam s₁) = infer-Lam m Γ s₁
  infer m Γ (App s₁ s₂) = infer-App m Γ s₁ s₂

  infer-Unit : (m : ℕ) → {n : ℕ} → (Γ : Cxt {m} n) →
         untypable Γ Unit
       ⊎ Σ[ m′′ ∈ ℕ ] Σ[ m′ ∈ ℕ ] Σ[ m≤m′′ ∈ m ≤ m′′ ]
         Σ[ σ ∈ AList TypeF m′′ m′ ] Σ[ t ∈ Type m′ ]
         Σ[ w ∈ WellTyped (applySubCxt≤ σ m≤m′′ Γ) t ]
         (erase w Unit × mgt σ Γ Unit t)
  infer-Unit m Γ =
    inj₂ (m , m , m≤m , anil , TUnit , Unit , Unit , mg-Unit)
    where
      m≤m : m ≤ m
      m≤m = n≤m+n 0 m

  infer-Var : (m : ℕ) → {n : ℕ} → (Γ : Cxt {m} n) → (x : Fin n) →
         untypable Γ (Var x)
       ⊎ Σ[ m′′ ∈ ℕ ] Σ[ m′ ∈ ℕ ] Σ[ m≤m′′ ∈ m ≤ m′′ ]
         Σ[ σ ∈ AList TypeF m′′ m′ ] Σ[ t ∈ Type m′ ]
         Σ[ w ∈ WellTyped (applySubCxt≤ σ m≤m′′ Γ) t ]
         (erase w (Var x) × mgt σ Γ (Var x) t)
  infer-Var m Γ x =
    inj₂ (m , m , m≤m , anil , t , VarX , eq , mg-Var x)
    where
      t : Type m
      t = lookup x Γ
      m≤m : m ≤ m
      m≤m = n≤m+n 0 m
      VarX : WellTyped (applySubCxt≤ anil m≤m Γ) t
      VarX rewrite applySubCxt≤Anil Γ m≤m = Var x
      eq : erase VarX (Var x)
      eq rewrite applySubCxt≤Anil Γ m≤m = Var x

  infer-Lam : (m : ℕ) → {n : ℕ} →
         (Γ : Cxt {m} n) → (s₁ : WellScoped (1 + n)) →
         untypable Γ (Lam s₁)
       ⊎ Σ[ m′′ ∈ ℕ ] Σ[ m′ ∈ ℕ ] Σ[ m≤m′′ ∈ m ≤ m′′ ]
         Σ[ σ ∈ AList TypeF m′′ m′ ] Σ[ t ∈ Type m′ ]
         Σ[ w ∈ WellTyped (applySubCxt≤ σ m≤m′′ Γ) t ]
         (erase w (Lam s₁) × mgt σ Γ (Lam s₁) t)
  infer-Lam m {n} Γ s₁
    with let
             t₂ : Type (1 + m)
             t₂ = ⟪ fromℕ m ⟫ -- new type variable
             Γ′ : Cxt {1 + m} n
             Γ′ = liftCxt 1 Γ
         in infer (1 + m) (t₂ ∷ Γ′) s₁
  ... | inj₁ ill-s₁ = -- s₁ is ill-typed
    inj₁ (illtyped-Lam Γ s₁ ill-s₁)
  ... | inj₂ (suc m′′ , m′ , s≤s m≤m′′ , σ , t₁ , w₁ , eraseW₁≡S₁ , mgσ) =
    inj₂ (suc m′′ , m′ , m≤1+m′′ , σ , σt₂ ⇒ t₁ ,
         LamW₁ , eraseLamW₁≡LamS₁ ,
         mg-Lam s₁ t₁ m≤m′′ σ w₁ eraseW₁≡S₁ mgσ)
    where
      σt₂ : Type m′
      σt₂ = applySub≤ σ (s≤s m≤m′′) t₂
        where
          t₂ : Type (1 + m)
          t₂ = ⟪ fromℕ m ⟫ -- new type variable
      m≤1+m′′ : m ≤ suc m′′
      m≤1+m′′ = ≤-step m≤m′′
      Γ′ : Cxt {1 + m} n -- the same as above
      Γ′ = liftCxt 1 Γ
      σΓ′≡σΓ : applySubCxt≤ σ (s≤s m≤m′′) Γ′ ≡
                  applySubCxt≤ σ m≤1+m′′ Γ
      σΓ′≡σΓ = subCxt≤+1 Γ (s≤s m≤m′′) m≤1+m′′ σ
      LamW₁ : WellTyped (applySubCxt≤ σ m≤1+m′′ Γ) (σt₂ ⇒ t₁)
      LamW₁ rewrite sym σΓ′≡σΓ = Lam σt₂ w₁
      eraseLamW₁≡LamS₁ : erase LamW₁ (Lam s₁)
      eraseLamW₁≡LamS₁ rewrite sym σΓ′≡σΓ
        = Lam w₁ s₁ eraseW₁≡S₁

  infer-App : (m : ℕ) → {n : ℕ} →
         (Γ : Cxt {m} n) → (s₁ s₂ : WellScoped n) →
         untypable Γ (App s₁ s₂)
       ⊎ Σ[ m′′ ∈ ℕ ] Σ[ m′ ∈ ℕ ] Σ[ m≤m′′ ∈ m ≤ m′′ ]
         Σ[ σ ∈ AList TypeF m′′ m′ ] Σ[ t ∈ Type m′ ]
         Σ[ w ∈ WellTyped (applySubCxt≤ σ m≤m′′ Γ) t ]
         (erase w (App s₁ s₂) × mgt σ Γ (App s₁ s₂) t)
  infer-App m {n} Γ s₁ s₂
    with infer m Γ s₁
  ... | inj₁ ill-s₁ = -- s₁ is ill-typed
    inj₁ (illtyped-App₁ Γ s₁ s₂ ill-s₁)
  ... | inj₂ (m₁′′ , m₁′ , m≤m₁′′ , σ₁ , t₁ , w₁ , eraseW₁≡S₁ , mgσ₁)
    with let
           σ₁Γ : Cxt n
           σ₁Γ = applySubCxt≤ σ₁ m≤m₁′′ Γ
         in infer m₁′ σ₁Γ s₂
  ... | inj₁ ill-s₂ = -- s₂ is ill-typed
    inj₁ (illtyped-App₂ Γ s₁ s₂ t₁ m≤m₁′′ σ₁ mgσ₁ ill-s₂)
  ... | inj₂ (m₂′′ , m₂′ , m₁′≤m₂′′ , σ₂ , t₂ , w₂ , eraseW₂≡S₂ , mgσ₂)
    with let
           t : Type (suc m₂′)
           t = ⟪ fromℕ m₂′ ⟫ -- new type variable
           m₂′≤1+m₂′ : m₂′ ≤ suc m₂′
           m₂′≤1+m₂′ = n≤m+n 1 m₂′
           σ₂t₁ : Type m₂′
           σ₂t₁ = applySub≤ σ₂ m₁′≤m₂′′ t₁ 
         in mgu (lift≤ m₂′≤1+m₂′ σ₂t₁) (lift≤ m₂′≤1+m₂′ t₂ ⇒ t)
  ... | inj₁ ill-unify = inj₁ (illtyped-unify Γ s₁ s₂ t₁ m≤m₁′′ m₁′≤m₂′′
                              σ₁ mgσ₁ σ₂ t₂ mgσ₂ ill-unify)
  ... | inj₂ (m₃′ , σ₃ , mgσ₃)
    = let
           m₂′′′ = proj₁ (σ→// σ₁ m₁′≤m₂′′)
           m₁′′≤m₂′′′ = proj₁ (proj₂ (σ→// σ₁ m₁′≤m₂′′))
           m₁′′≤m₂′′′//m₁′≤m₂′′ = proj₂ (proj₂ (σ→// σ₁ m₁′≤m₂′′))
      in let
           σ₂₁ : AList TypeF m₂′′′ m₂′
           σ₂₁ = σ₂ +⟨ m₁′′≤m₂′′′//m₁′≤m₂′′ ⟩ σ₁
           m₂′≤1+m₂′ : m₂′ ≤ suc m₂′
           m₂′≤1+m₂′ = n≤m+n 1 m₂′
           1+m₂′′′ = proj₁ (σ→// σ₂₁ m₂′≤1+m₂′)
           m₂′′′≤1+m₂′′′ = proj₁ (proj₂ (σ→// σ₂₁ m₂′≤1+m₂′))
           m₂′′′≤1+m₂′′′//m₂′≤1+m₂′ = proj₂ (proj₂ (σ→// σ₂₁ m₂′≤1+m₂′))
    in inj₂ (1+m₂′′′ , m₃′ , m≤1+m₂′′′ , σ₃₂₁ , σ₃t ,
            AppW₁W₂ , eraseAppW₁W₂≡AppS₁S₂ ,
            mg-App s₁ m≤m₁′′ σ₁ t₁ w₁ eraseW₁≡S₁ mgσ₁
                   s₂ m₁′≤m₂′′ σ₂ t₂ w₂ eraseW₂≡S₂ mgσ₂
                   σ₃ mgσ₃)
    where
      σ₃t : Type m₃′
      σ₃t = applySub σ₃ t
        where
           t : Type (suc m₂′)
           t = ⟪ fromℕ m₂′ ⟫ -- new type variable
      m₂′′′ = proj₁ (σ→// σ₁ m₁′≤m₂′′)
      m₁′′≤m₂′′′ = proj₁ (proj₂ (σ→// σ₁ m₁′≤m₂′′))
      m₁′′≤m₂′′′//m₁′≤m₂′′ = proj₂ (proj₂ (σ→// σ₁ m₁′≤m₂′′))
      m₂′≤1+m₂′ = n≤m+n 1 m₂′
      σ₂₁ : AList TypeF m₂′′′ m₂′
      σ₂₁ = σ₂ +⟨ m₁′′≤m₂′′′//m₁′≤m₂′′ ⟩ σ₁
      1+m₂′′′ = proj₁ (σ→// σ₂₁ m₂′≤1+m₂′)
      m₂′′′≤1+m₂′′′ = proj₁ (proj₂ (σ→// σ₂₁ m₂′≤1+m₂′))
      m₂′′′≤1+m₂′′′//m₂′≤1+m₂′ = proj₂ (proj₂ (σ→// σ₂₁ m₂′≤1+m₂′))
      -- t = ⟪ fromℕ m₂′ ⟫
      -- m₂′≤1+m₂′ = n≤m+n 1 m₂′
      -- σ₃t : Type m₃′
      -- σ₃t = applySub σ₃ t
      -- w₁ = σ₁Γ |- s₁ : t₁
      σ₁Γ = applySubCxt≤ σ₁ m≤m₁′′ Γ
      -- σ₂w₁ = σ₂σ₁Γ |- s₁ : σ₂t₁
      σ₂w₁ = applySubWTerm≤ σ₂ m₁′≤m₂′′ w₁
      σ₂σ₁Γ = applySubCxt≤ σ₂ m₁′≤m₂′′ σ₁Γ
      σ₂t₁ = applySub≤ σ₂ m₁′≤m₂′′ t₁ 
      -- σ₃σ₂w₁ = σ₃σ₂σ₁Γ |- s₁ : σ₃σ₂t₁
      σ₃σ₂w₁ = applySubWTerm≤ σ₃ m₂′≤1+m₂′ σ₂w₁
      σ₃σ₂σ₁Γ = applySubCxt≤ σ₃ m₂′≤1+m₂′ σ₂σ₁Γ
      σ₃σ₂t₁ = applySub≤ σ₃ m₂′≤1+m₂′ σ₂t₁

      -- w₂ = σ₂σ₁Γ |- s₂ : t₂
      -- σ₃w₂ = σ₃σ₂σ₁Γ |- s₂ : σ₃t₂
      σ₃w₂ = applySubWTerm≤ σ₃ m₂′≤1+m₂′ w₂
      σ₃t₂ = applySub≤ σ₃ m₂′≤1+m₂′ t₂

      -- σ₃σ₂t₁ ≡ σ₃t₂ ⇒ σ₃t
      -- σ₃t : Type m₃′
      -- σ₃t = applySub σ₃ t
      σ₃σ₂t₁≡σ₃t₂⇒σ₃t : σ₃σ₂t₁ ≡ σ₃t₂ ⇒ σ₃t
      σ₃σ₂t₁≡σ₃t₂⇒σ₃t = proj₁ mgσ₃
      σ₃t₂⇒σ₃t≡σ₃σ₂t₁ : σ₃t₂ ⇒ σ₃t ≡ σ₃σ₂t₁
      σ₃t₂⇒σ₃t≡σ₃σ₂t₁ = sym σ₃σ₂t₁≡σ₃t₂⇒σ₃t

      -- σ₃σ₂W₁ = σ₃σ₂σ₁Γ |- s₁ : σ₃t₂ ⇒ σ₃t
      σ₃σ₂W₁ : WellTyped σ₃σ₂σ₁Γ (σ₃t₂ ⇒ σ₃t)
      σ₃σ₂W₁ rewrite σ₃t₂⇒σ₃t≡σ₃σ₂t₁ = σ₃σ₂w₁

      -- substitutions
      -- σ₂₁ : AList TypeF m₂′′′ m₂′
      -- σ₂₁ = σ₂ +⟨ m₁′′≤m₂′′′//m₁′≤m₂′′ ⟩ σ₁
      m≤m₂′′′ : m ≤ m₂′′′
      m≤m₂′′′ = ≤-trans m≤m₁′′ m₁′′≤m₂′′′
      σ₃₂₁ : AList TypeF 1+m₂′′′ m₃′
      σ₃₂₁ = σ₃ +⟨ m₂′′′≤1+m₂′′′//m₂′≤1+m₂′ ⟩ σ₂₁
      m≤1+m₂′′′ : m ≤ 1+m₂′′′
      m≤1+m₂′′′ = ≤-trans m≤m₂′′′ m₂′′′≤1+m₂′′′

      -- environments
      σ₂₁Γ : Cxt n
      σ₂₁Γ = applySubCxt≤ σ₂₁ m≤m₂′′′ Γ
      σ₂₁Γ≡σ₂σ₁Γ : σ₂₁Γ ≡ σ₂σ₁Γ
      σ₂₁Γ≡σ₂σ₁Γ =
        begin
          applySubCxt≤ σ₂₁ m≤m₂′′′ Γ
        ≡⟨ subCxtTrans Γ σ₁ m₁′≤m₂′′ m₁′′≤m₂′′′ m₁′′≤m₂′′′//m₁′≤m₂′′ σ₂ σ₂₁ m≤m₁′′ m≤m₂′′′ refl ⟩
          applySubCxt≤ σ₂ m₁′≤m₂′′ σ₁Γ
        ∎ where open ≡-Reasoning
      σ₃₂₁Γ : Cxt n
      σ₃₂₁Γ = applySubCxt≤ σ₃₂₁ m≤1+m₂′′′ Γ
      σ₃₂₁Γ≡σ₃σ₂σ₁Γ : σ₃₂₁Γ ≡ σ₃σ₂σ₁Γ
      σ₃₂₁Γ≡σ₃σ₂σ₁Γ =
        begin
          applySubCxt≤ σ₃₂₁ m≤1+m₂′′′ Γ
        ≡⟨ subCxtTrans Γ σ₂₁ m₂′≤1+m₂′ m₂′′′≤1+m₂′′′ m₂′′′≤1+m₂′′′//m₂′≤1+m₂′ σ₃ σ₃₂₁ m≤m₂′′′ m≤1+m₂′′′ refl ⟩
          applySubCxt≤ σ₃ m₂′≤1+m₂′ σ₂₁Γ
        ≡⟨ cong (applySubCxt≤ σ₃ m₂′≤1+m₂′) σ₂₁Γ≡σ₂σ₁Γ ⟩
          applySubCxt≤ σ₃ m₂′≤1+m₂′ σ₂σ₁Γ
        ∎ where open ≡-Reasoning

      -- AppW₁W₂
      W₁ : WellTyped σ₃₂₁Γ (σ₃t₂ ⇒ σ₃t)
      W₁ rewrite σ₃₂₁Γ≡σ₃σ₂σ₁Γ = σ₃σ₂W₁
      W₂ : WellTyped σ₃₂₁Γ σ₃t₂
      W₂ rewrite σ₃₂₁Γ≡σ₃σ₂σ₁Γ = σ₃w₂
      AppW₁W₂ : WellTyped σ₃₂₁Γ σ₃t
      AppW₁W₂ = App W₁ W₂

      -- erase
      eraseAppW₁W₂≡AppS₁S₂ : erase AppW₁W₂ (App s₁ s₂)
      eraseAppW₁W₂≡AppS₁S₂ = App W₁ s₁ W₂ s₂ eraseW₁≡S₁′ eraseW₂≡S₂′
        where eraseW₁≡S₁′ : erase W₁ s₁
              eraseW₁≡S₁′ rewrite σ₃₂₁Γ≡σ₃σ₂σ₁Γ | σ₃t₂⇒σ₃t≡σ₃σ₂t₁
                = eraseSubWTerm≤ σ₃ m₂′≤1+m₂′ σ₂w₁ s₁
                   (eraseSubWTerm≤ σ₂ m₁′≤m₂′′ w₁ s₁ eraseW₁≡S₁)
              eraseW₂≡S₂′ : erase W₂ s₂
              eraseW₂≡S₂′ rewrite σ₃₂₁Γ≡σ₃σ₂σ₁Γ | σ₃t₂⇒σ₃t≡σ₃σ₂t₁
                = eraseSubWTerm≤ σ₃ m₂′≤1+m₂′ w₂ s₂ eraseW₂≡S₂
